﻿using System.Buffers;
using System.Runtime.Intrinsics;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Json;

namespace histcopy
{
    public class Histcopy
    {
        #region 変数
        // 必須設定
        private readonly string SourceDirectoryName;
        private readonly string RootDirectoryName;
        private readonly string RecentDirectoryName;
        private readonly string HistoryDirectoryName;
        
        // 任意設定
        public double VerifyCycle { get; init; } = 100.0; // 日数

        // 制御
        private readonly CancellationToken Token;
        private readonly CountdownEvent CountDown = new(1);
        private readonly Dictionary<FileKey, List<FileInfo>> Moves = [];
        private readonly List<FileInfo?> Copies = [];
        private StreamWriter? LogWriter;

        // 統計
        private int  CopyCount,  VerifyCount,  RecoverCount,  ErrorCount;
        private long CopyLength, VerifyLength, RecoverLength, FileLength;
        #endregion

        #region 進捗
        public event EventHandler<ProgressEventArgs>? Progress;

        public class ProgressEventArgs(string message) : EventArgs
        {
            public string Message { get; } = message;
        }
        #endregion

        #region コンストラクタ
        public Histcopy(string source, string destination, CancellationToken token)
        {
            SourceDirectoryName = (source.StartsWith(@"\\") ? "" : @"\\?\") + source;
            RootDirectoryName = (destination.StartsWith(@"\\") ? "" : @"\\?\") + destination;
            RecentDirectoryName = Path.Combine(RootDirectoryName, "recent");
            HistoryDirectoryName = Path.Combine(RootDirectoryName, $"{DateTime.Now:yyyy-MM-dd}");
            Token = token;
        }
        #endregion

        #region 実行
        public void Run()
        {
            // ソース存在チェック
            var sourceDirectory = new DirectoryInfo(SourceDirectoryName);
            if (!sourceDirectory.Exists)
                throw new DirectoryNotFoundException($"{SourceDirectoryName} は見つかりません。");

            // ログ作成
            var historyDirectory = new DirectoryInfo(HistoryDirectoryName);
            if (!historyDirectory.Exists)
                historyDirectory.Create();
            using var logWriter = new StreamWriter(Path.Combine(HistoryDirectoryName, "histcopy_log.txt"), true);
            LogWriter = logWriter;
            WriteLog($"-- {DateTime.Now:HH:mm:ss} --");

            // ディレクトリスキャン、移動
            Task.Run(() => Scan(sourceDirectory));
            CountDown.Wait();

            // ファイル移動（リサイクル）
            Parallel.For(0, Copies.Count, copyIndex =>
            {
                try
                {
                    var sourceFile = Copies[copyIndex];
                    if (sourceFile == null || Token.IsCancellationRequested)
                        return;
                    var relativeFileName = GetRelativeName(sourceFile, SourceDirectoryName);
                    var recentFileName = Path.Combine(RecentDirectoryName, relativeFileName);
                    if (Moves.TryGetValue(new FileKey(sourceFile.LastWriteTime, sourceFile.Length), out var list))
                    {
                        lock (list)
                        {
                            var foundIndex = -1;
                            for (var index = 0; index < list.Count; index++)
                            {
                                if (list[index].Name == sourceFile.Name)
                                {
                                    foundIndex = index;
                                    break;
                                }
                            }
                            if (foundIndex < 0)
                            {
                                for (var index = 0; index < list.Count; index++)
                                {
                                    if (CompareHeaderFooter(list[index], sourceFile))
                                    {
                                        foundIndex = index;
                                        break;
                                    }
                                }
                            }
                            if (foundIndex >= 0)
                            {
                                var movedFile = list[foundIndex];
                                movedFile.MoveTo(recentFileName);
                                list.RemoveAt(foundIndex);
                                Copies[copyIndex] = null;
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    WriteLog(ex.Message);
                    Interlocked.Increment(ref ErrorCount);
                }
            });
            Moves.Clear();
            Moves.TrimExcess();

            // 履歴にある空のディレクトリを削除
            CountDown.Reset();
            Task.Run(() => DeleteEmptyDirectory(historyDirectory, null));
            CountDown.Wait();
            CountDown.Dispose();

            // ファイルコピー
            Copies.Sort((file1, file2) => (file2 == null ? "" : file2.FullName).CompareTo(file1 == null ? "" : file1.FullName));
            for (var copyIndex = Copies.Count - 1; copyIndex >= 0 && !Token.IsCancellationRequested; copyIndex--)
            {
                try
                {
                    var sourceFile = Copies[copyIndex];
                    Copies.RemoveAt(copyIndex);
                    if (sourceFile != null)
                    {
                        var relativeFileName = GetRelativeName(sourceFile, SourceDirectoryName);
                        var recentFileName = Path.Combine(RecentDirectoryName, relativeFileName);
                        sourceFile.CopyTo(recentFileName);
                        WriteLog(relativeFileName);
                        CopyCount++;
                        CopyLength += sourceFile.Length;
                    }
                }
                catch (Exception ex)
                {
                    WriteLog(ex.Message);
                    ErrorCount++;
                }
            }
            Copies.Clear();
            Copies.TrimExcess();

            // ベリファイ
            using (var verifyState = new VerifyState(this, Path.Combine(RootDirectoryName, "verify.json"), SourceDirectoryName))
            {
                var rate = Math.Max(0.0, Math.Min(1.0, (verifyState.ExecutionDate - verifyState.LastExecutionDate).TotalDays / VerifyCycle));
                var leftLength = (long)Math.Ceiling(FileLength * rate);
                while (!Token.IsCancellationRequested && leftLength > 0 && verifyState.Read())
                {
                    try
                    {
                        // ファイル取得
                        var sourceFile = verifyState.File ?? throw new NullReferenceException();
                        var relativeFileName = GetRelativeName(sourceFile, SourceDirectoryName);
                        var recentFile = new FileInfo(Path.Combine(RecentDirectoryName, relativeFileName));

                        // ファイルオープン、バッファ確保
                        var isDataEquals = true;
                        using (var sourceStream = sourceFile.OpenRead())
                        using (var recentStream = recentFile.OpenRead())
                        using (var buffer1 = new VerifyBufferSet())
                        using (var buffer2 = new VerifyBufferSet())
                        {
                            VerifyCount++;
                            sourceStream.Seek(verifyState.Position, SeekOrigin.Begin);
                            recentStream.Seek(verifyState.Position, SeekOrigin.Begin);

                            var buffer = buffer1;
                            var lastBuffer = buffer2;
                            var isEndOfFile = false;
                            while (!Token.IsCancellationRequested && leftLength > 0 && isDataEquals && !isEndOfFile)
                            {
                                // 読み込み、比較
                                buffer.BeginRead(sourceStream, recentStream);
                                isDataEquals = lastBuffer.Compare();
                                buffer.EndRead();

                                // 統計値に加算
                                VerifyLength += buffer.Buffer1.ReadLength;
                                leftLength -= buffer.Buffer1.ReadLength;
                                verifyState.Position += buffer.Buffer1.ReadLength;
                                isEndOfFile = buffer.Buffer1.ReadLength == 0 && buffer.Buffer2.ReadLength == 0;

                                // バッファを入れ替え
                                (lastBuffer, buffer) = (buffer, lastBuffer);
                            }
                            if (isDataEquals)
                                isDataEquals = lastBuffer.Compare();
                        }

                        // データ不一致のとき、再コピー
                        if (!isDataEquals)
                        {
                            // 履歴ディレクトリを作成、ユニークな名前を生成・追加
                            var relativeDirectoryName = GetRelativeName(sourceFile.Directory ?? throw new NullReferenceException(), SourceDirectoryName);
                            var historyDirectory2 = new HistoryDirectory(this, Path.Combine(HistoryDirectoryName, relativeDirectoryName));
                            var uniqueName = historyDirectory2.CreateAndAdd(recentFile.Name);

                            // 最近ファイルを移動
                            File.Move(recentFile.FullName, Path.Combine(historyDirectory2.FullName, uniqueName));

                            // コピー
                            sourceFile.CopyTo(recentFile.FullName);
                            WriteLog(relativeFileName);
                            RecoverCount++;
                            RecoverLength += sourceFile.Length;
                        }
                    }
                    catch (Exception ex)
                    {
                        WriteLog(ex.Message);
                        ErrorCount++;
                    }
                }
            }

            // 結果表示
            WriteLog($"-- {DateTime.Now:HH:mm:ss} --");
            WriteLog("copy", CopyCount, CopyLength);
            WriteLog("verify", VerifyCount, VerifyLength);
            WriteLog("recover", RecoverCount, RecoverLength);
            WriteLog("error", ErrorCount);
            WriteLog("");
        }
        #endregion

        #region ソースディレクトリをスキャン
        private void Scan(DirectoryInfo sourceDirectory)
        {
            try
            {
                // キャンセルされたときは抜ける
                if (Token.IsCancellationRequested)
                    return;

                // ディレクトリを取得
                var relativeDirectoryName = GetRelativeName(sourceDirectory, SourceDirectoryName);
                var recentDirectory = new DirectoryInfo(Path.Combine(RecentDirectoryName, relativeDirectoryName));
                var historyDirectory = new HistoryDirectory(this, Path.Combine(HistoryDirectoryName, relativeDirectoryName));

                // 最近ディレクトリが存在しない場合は作成
                var recentDirectoryExists = recentDirectory.Exists;
                if (!recentDirectoryExists)
                    recentDirectory.Create();

                // ディレクトリ情報を取得
                var sourceInfos = sourceDirectory.GetFileSystemInfos().ToDictionary(info => info.Name.ToUpper());
                var recentInfos = recentDirectoryExists ? recentDirectory.GetFileSystemInfos().ToDictionary(info => info.Name.ToUpper()) : [];

                // ソースを処理
                foreach (var sourceInfo in sourceInfos.Values)
                {
                    if (sourceInfo is DirectoryInfo sourceSubDirectory)
                    {
                        // ソースがサブディレクトリとき、スキャン
                        CountDown.AddCount(1);
                        Task.Run(() => Scan(sourceSubDirectory));
                    }
                    else
                    {
                        // ソースがファイルのとき
                        var sourceFile = (FileInfo)sourceInfo;
                        Interlocked.Add(ref FileLength, sourceFile.Length);
                        if (!recentInfos.TryGetValue(sourceFile.Name.ToUpper(), out FileSystemInfo? recentInfo))
                        {
                            // 最近にファイル・ディレクトリが存在しないとき、ファイルコピー
                            lock (Copies)
                                Copies.Add(sourceFile);
                        }
                        else
                        {
                            if (recentInfo is DirectoryInfo recentSubDirectory)
                            {
                                // 最近にディレクトリが存在するとき、ディレクトリ移動、ファイルコピー
                                AddMoveAndCopy(recentSubDirectory, historyDirectory, sourceFile);
                            }
                            else
                            {
                                // 最近にファイルが存在し、ソースファイルの属性と一致しないとき、ファイル移動、ファイルコピー
                                var recentFile = (FileInfo)recentInfo;
                                if (sourceFile.Name != recentFile.Name ||
                                    sourceFile.Length != recentFile.Length ||
                                    sourceFile.LastWriteTime != recentFile.LastWriteTime)
                                {
                                    AddMoveAndCopy(recentFile, historyDirectory, sourceFile);
                                }
                            }
                        }
                    }
                }

                // 最近にファイル・ディレクトリが存在し、ソースにファイル・ディレクトリが存在しないとき、移動
                foreach (var recentInfo in recentInfos.Values)
                    if (!sourceInfos.ContainsKey(recentInfo.Name.ToUpper()))
                        AddMoveAndCopy(recentInfo, historyDirectory, null);
            }
            catch (Exception ex)
            {
                WriteLog(ex.Message);
                Interlocked.Increment(ref ErrorCount);
            }
            finally
            {
                CountDown.Signal();
            }
        }
        #endregion

        #region ファイル・ディレクトリを移動、ファイルコピー
        private void AddMoveAndCopy(FileSystemInfo recentInfo, HistoryDirectory historyDirectory, FileInfo? sourceFile)
        {
            CountDown.AddCount(1);
            Task.Run(() => {
                try
                {
                    // キャンセルされたときは抜ける
                    if (Token.IsCancellationRequested)
                        return;

                    // 履歴ディレクトリを作成、ユニークな名前を生成・追加
                    var uniqueName = historyDirectory.CreateAndAdd(recentInfo.Name);
                    var destination = Path.Combine(historyDirectory.FullName, uniqueName);
                    var relativeName = GetRelativeName(recentInfo, RecentDirectoryName);

                    if (recentInfo is DirectoryInfo recentSubDirectory)
                    {
                        // 最近ディレクトリを移動
                        recentSubDirectory.MoveTo(destination);

                        // ディレクトリを再帰スキャンし、ファイルを移動リストに追加
                        CountDown.AddCount(1);
                        Task.Run(() => AddMoves(recentSubDirectory));
                    }
                    else if (recentInfo is FileInfo recentFile)
                    {
                        // 最近ファイルを移動
                        recentFile.MoveTo(destination);

                        // ファイルを移動リストに追加
                        AddMoves(recentFile);
                    }

                    // コピーリストに追加
                    if (sourceFile != null)
                        lock (Copies)
                            Copies.Add(sourceFile);
                }
                catch (Exception ex)
                {
                    WriteLog(ex.Message);
                    Interlocked.Increment(ref ErrorCount);
                }
                finally
                {
                    CountDown.Signal();
                }
            });
        }
        #endregion

        #region 空ディレクトリを削除
        private void DeleteEmptyDirectory(DirectoryInfo directory, DirectoryItem? parent)
        {
            try
            {
                // キャンセルされたときは抜ける
                if (Token.IsCancellationRequested)
                    return;

                // ディレクトリ情報を読み取り、項目を生成
                var infos = directory.GetFileSystemInfos();
                var item = new DirectoryItem(directory, parent, infos.Length);

                if (item.Count == 0)
                {
                    // ディレクトリが空のときは上位ディレクトリも含めて削除
                    while (item.Count == 0)
                    {
                        item.Directory.Delete();
                        if (item.Parent == null)
                            break;
                        item = item.Parent;
                        Interlocked.Decrement(ref item.Count);
                    }
                }
                else
                {
                    foreach (var subDirectory in infos.OfType<DirectoryInfo>())
                    {
                        // サブディレクトリをスキャン
                        CountDown.AddCount(1);
                        Task.Run(() => DeleteEmptyDirectory(subDirectory, item));
                    }
                }
            }
            catch (Exception ex)
            {
                WriteLog(ex.Message);
                Interlocked.Increment(ref ErrorCount);
            }
            finally
            {
                CountDown.Signal();
            }
        }

        private class DirectoryItem(DirectoryInfo directory, DirectoryItem? parent, int count)
        {
            public readonly DirectoryInfo Directory = directory;
            public readonly DirectoryItem? Parent = parent;
            public int Count = count;
        }
        #endregion

        #region 移動ファイルを追加
        private void AddMoves(FileInfo file)
        {
            var key = new FileKey(file.LastWriteTime, file.Length);
            lock (Moves)
            {
                if (!Moves.TryGetValue(key, out var list))
                {
                    list = [];
                    Moves.Add(key, list);
                }
                list.Add(file);
            }
        }

        private void AddMoves(DirectoryInfo directory)
        {
            try
            {
                // キャンセルされたときは抜ける
                if (Token.IsCancellationRequested)
                    return;

                var infos = directory.GetFileSystemInfos();
                foreach (var info in infos)
                {
                    if (info is DirectoryInfo subDirectory)
                    {
                        // サブディレクトリをスキャン
                        CountDown.AddCount(1);
                        Task.Run(() => AddMoves(subDirectory));
                    }
                    else if (info is FileInfo file)
                    {
                        // 移動ファイルを追加
                        AddMoves(file);
                    }
                }
            }
            catch (Exception ex)
            {
                WriteLog(ex.Message);
                Interlocked.Increment(ref ErrorCount);
            }
            finally
            {
                CountDown.Signal();
            }
        }

        private record FileKey(DateTime Date, long Length);
        #endregion

        #region 履歴ディレクトリ
        private class HistoryDirectory(object parent, string fullName)
        {
            private readonly object Parent = parent;
            public readonly string FullName = fullName;
            private HashSet<string>? Keys;

            public string CreateAndAdd(string baseName)
            {
                lock (this)
                {
                    if (Keys == null)
                    {
                        var directory = new DirectoryInfo(FullName);
                        lock (Parent) // dir1 と dir1\dir2 の作成を全体で排他
                        {
                            if (!directory.Exists)
                            {
                                directory.Create();
                                Keys = [];
                            }
                        }
                        if (Keys == null)
                        {
                            var infos = directory.GetFileSystemInfos();
                            Keys = (from fsi in infos select fsi.Name.ToUpper()).ToHashSet();
                        }
                    }
                    var name = baseName;
                    var index = 0;
                    while (!Keys.Add(name.ToUpper()))
                    {
                        name = $"{baseName}.{index++}";
                    }
                    return name;
                }
            }
        }
        #endregion

        #region ヘッダ・フッタを比較
        private static bool CompareHeaderFooter(FileInfo file1, FileInfo file2)
        {
            // サイズチェック
            if (file1.Length != file2.Length)
                return false;

            if (file1.Length > 0)
            {
                const int headerLength = 4096;
                const int footerLength = 4096;

                // ファイルオープン
                using var stream1 = file1.OpenRead();
                using var stream2 = file2.OpenRead();

                // ヘッダ比較
                if (!CompareFilePart(stream1, stream2, 0, headerLength))
                    return false;

                // フッタ比較
                if (file1.Length > headerLength && !CompareFilePart(stream1, stream2, Math.Max(headerLength, file1.Length - footerLength), footerLength))
                    return false;
            }

            return true;
        }

        private static bool CompareFilePart(FileStream stream1, FileStream stream2, long offset, int length)
        {
            if (length == 0)
                return true;

            byte[]? buffer1 = null;
            byte[]? buffer2 = null;
            try
            {
                stream1.Seek(offset, SeekOrigin.Begin);
                buffer1 = ArrayPool<byte>.Shared.Rent(length);
                var readLength1 = stream1.Read(buffer1, 0, length);

                stream2.Seek(offset, SeekOrigin.Begin);
                buffer2 = ArrayPool<byte>.Shared.Rent(length);
                var readLength2 = stream2.Read(buffer2, 0, length);

                return CompareBuffer(buffer1, buffer2, readLength1, readLength2);
            }
            finally
            {
                if (buffer1 != null)
                    ArrayPool<byte>.Shared.Return(buffer1);
                if (buffer2 != null)
                    ArrayPool<byte>.Shared.Return(buffer2);
            }
        }
        #endregion

        #region バッファ内容を比較
        private static bool CompareBuffer(byte[] buffer1, byte[] buffer2, int length1, int length2)
        {
            if (length1 != length2)
            {
                return false;
            }
            else if (Vector512.IsHardwareAccelerated && length1 >= 64)
            {
                for (int index = 0, length64 = (length1 - 1) & 0x7fff_ffc0; index < length64; index += 64)
                    if (Vector512.Create(buffer1, index) != Vector512.Create(buffer2, index))
                        return false;
                return Vector512.Create(buffer1, length1 - 64) == Vector512.Create(buffer2, length1 - 64);
            }
            else if (Vector256.IsHardwareAccelerated && length1 >= 32)
            {
                for (int index = 0, length32 = (length1 - 1) & 0x7fff_ffe0; index < length32; index += 32)
                    if (Vector256.Create(buffer1, index) != Vector256.Create(buffer2, index))
                        return false;
                return Vector256.Create(buffer1, length1 - 32) == Vector256.Create(buffer2, length1 - 32);
            }
            else
            {
                for (var index = 0; index < length1; index++)
                    if (buffer1[index] != buffer2[index])
                        return false;
                return true;
            }
        }
        #endregion

        #region 相対名を取得
        private static string GetRelativeName(FileSystemInfo info, string directoryName)
        {
            var name = info.FullName;
            var index = directoryName.Length;
            if (!directoryName.EndsWith('\\'))
                index++;
            return index >= name.Length ? "" : name[index..];
        }
        #endregion

        #region ログに書き込み
        private void WriteLog(string message)
        {
            if (LogWriter != null)
                lock (LogWriter)
                    try { LogWriter.WriteLine(message); } catch { }
            if (message.Length > 0)
                try { Progress?.Invoke(this, new ProgressEventArgs(message)); } catch { }
        }

        private void WriteLog(string message, int count)
        {
            if (count != 0)
            {
                var unit = count >= 2 ? "files" : "file";
                WriteLog($"{message} : {count:N0} {unit}");
            }
        }

        private void WriteLog(string message, int count, long length)
        {
            if (count != 0)
            {
                var countUnit = count >= 2 ? "files" : "file";
                var value = (double)length;
                var index = (int)Math.Floor(Math.Max(1.0, Math.Log(value, 1024.0)));
                var lengthUnit = string.Concat("KMGTPEZ".AsSpan(index - 1, 1), "B");
                value /= Math.Pow(1024.0, index);
                var strLength = value.ToString(value >= 100.0 ? "N0" : value >= 10.0 ? "N1" : "N2");
                if (strLength.EndsWith(".00"))
                    strLength = strLength[..^3];
                else if (strLength.EndsWith(".0"))
                    strLength = strLength[..^2];
                WriteLog($"{message} : {count:N0} {countUnit}, {strLength} {lengthUnit}");
            }
        }
        #endregion
                
        #region ベリファイ状態
        private class VerifyState : IDisposable
        {
            private readonly Histcopy Histcopy;
            private readonly string JsonFileName;
            private readonly string DirectoryName;
            private DirectoryItem Directory;
            private bool IsResumed;
            private long NextPosition;
            public readonly DateTime ExecutionDate;
            public readonly DateTime LastExecutionDate;
            public FileInfo? File;
            public long Position;

            public VerifyState(Histcopy histcopy, string jsonFileName, string directoryName)
            {
                Histcopy = histcopy;
                JsonFileName = jsonFileName;
                DirectoryName = directoryName;
                ExecutionDate = DateTime.Now;
                Directory = new DirectoryItem(new DirectoryInfo(DirectoryName), null);
                Reset(false);

                try
                {
                    // JSONファイルを読み込み
                    var file = new FileInfo(JsonFileName);
                    if (!file.Exists)
                        return;
                    JsonData data;
                    using (var stream = file.OpenRead())
                    {
                        var serializer = new DataContractJsonSerializer(typeof(JsonData));
                        data = (JsonData)(serializer.ReadObject(stream) ?? throw new NullReferenceException());
                    }
                    LastExecutionDate = data.ExecutionDate;

                    // パスを分解し、名前キューを作成
                    var nameQueue = new Queue<string>(data.FileName.Split('\\'));
                    var name = nameQueue.Dequeue();
                    while (Directory.Queue.Count > 0)
                    {
                        var info = Directory.Queue.Peek();
                        var result = info.Name.CompareTo(name);

                        // ディレクトリのとき、次の階層へ移動する
                        if (result == 0 && nameQueue.Count > 0 && info is DirectoryInfo subDirectory)
                        {
                            Directory.Queue.Dequeue();
                            Directory = new DirectoryItem(subDirectory, Directory);
                            name = nameQueue.Dequeue();
                            continue;
                        }

                        // ファイルが見つかったとき、または見つからないときは抜ける
                        if (result >= 0)
                        {
                            NextPosition = result == 0 ? data.Position : 0;
                            break;
                        }

                        Directory.Queue.Dequeue();
                    }

                    IsResumed = true;
                }
                catch (Exception ex)
                {
                    Histcopy.WriteLog(ex.Message);
                    Histcopy.ErrorCount++;
                    Reset();
                }
                finally
                {
                    if (LastExecutionDate == DateTime.MinValue)
                        LastExecutionDate = ExecutionDate.AddDays(-1.0);
                }
            }

            private void Reset(bool isDirectorySet = true)
            {
                if (isDirectorySet)
                    Directory = new DirectoryItem(new DirectoryInfo(DirectoryName), null);
                IsResumed = false;
                NextPosition = 0;
                File = null;
                Position = 0;
            }

            public bool Read()
            {
                while (true)
                {
                    if (Directory.Queue.Count > 0)
                    {
                        var info = Directory.Queue.Dequeue();
                        if (info is FileInfo file)
                        {
                            // キュー項目がファイルのとき、カレント項目をセット、true（データあり）を返す
                            File = file;
                            Position = NextPosition;
                            NextPosition = 0;
                            return true;
                        }
                        try
                        {
                            // キュー項目がディレクトリのとき、次の階層へ移動する。読み取れないときは次の項目へ移動する。
                            Directory = new DirectoryItem((DirectoryInfo)info, Directory);
                        }
                        catch (Exception ex)
                        {
                            Histcopy.WriteLog(ex.Message);
                            Histcopy.ErrorCount++;
                        }
                        continue;
                    }

                    // 最後に到達したとき
                    if (Directory.Parent == null)
                    {
                        // 途中から読み始めたときは最初から読み直し
                        if (IsResumed)
                        {
                            Reset();
                            continue;
                        }
                        // 全体を一巡しているときは false（データなし）を返す
                        File = null;
                        Position = 0;
                        NextPosition = 0;
                        return false;
                    }

                    // 上の階層へ移動
                    Directory = Directory.Parent;
                }
            }

            public void Dispose()
            {
                try
                {
                    var data = new JsonData()
                    {
                        FileName = File == null ? "" : GetRelativeName(File, DirectoryName),
                        Position = Position,
                        ExecutionDate = ExecutionDate
                    };
                    using var stream = new FileStream(JsonFileName, FileMode.Create);
                    var serializer = new DataContractJsonSerializer(typeof(JsonData));
                    serializer.WriteObject(stream, data);
                }
                catch (Exception ex)
                {
                    Histcopy.WriteLog(ex.Message);
                    Histcopy.ErrorCount++;
                }
            }

            [DataContract]
            private class JsonData
            {
                [DataMember] public required string FileName;
                [DataMember] public long Position;
                [DataMember] public DateTime ExecutionDate;
            }

            private class DirectoryItem(DirectoryInfo directory, DirectoryItem? parent)
            {
                public readonly DirectoryInfo Directory = directory;
                public readonly DirectoryItem? Parent = parent;
                public readonly Queue<FileSystemInfo> Queue = new(from info in directory.GetFileSystemInfos() orderby info.Name select info);
            }
        }
        #endregion

        #region ベリファイバッファセット
        private class VerifyBufferSet : IDisposable
        {
            public readonly BufferInfo Buffer1 = new();
            public readonly BufferInfo Buffer2 = new();
            private Task? ReadTask;

            public void BeginRead(Stream stream1, Stream stream2)
            {
                if (ReadTask != null)
                    throw new InvalidOperationException();
                ReadTask = Task.WhenAll(Buffer1.ReadAsync(stream1), Buffer2.ReadAsync(stream2));
            }

            public void EndRead()
            {
                if (ReadTask != null)
                {
                    ReadTask.Wait();
                    ReadTask.Dispose();
                    ReadTask = null;
                }
            }

            public bool Compare()
            {
                return CompareBuffer(Buffer1.Bytes, Buffer2.Bytes, Buffer1.ReadLength, Buffer2.ReadLength);
            }

            public void Dispose()
            {
                EndRead();
                Buffer1.Dispose();
                Buffer2.Dispose();
            }

            public sealed class BufferInfo : IDisposable
            {
                public const int BufferLength = 1048576;
                public readonly byte[] Bytes = ArrayPool<byte>.Shared.Rent(BufferLength);
                public int ReadLength;

                public async Task ReadAsync(Stream stream)
                {
                    ReadLength = await stream.ReadAsync(Bytes.AsMemory(0, BufferLength), CancellationToken.None);
                }

                public void Dispose()
                {
                    ArrayPool<byte>.Shared.Return(Bytes);
                }
            }
        }
        #endregion
    }
}
